;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: snark -*-
;;; File: output.lisp
;;; The contents of this file are subject to the Mozilla Public License
;;; Version 1.1 (the "License"); you may not use this file except in
;;; compliance with the License. You may obtain a copy of the License at
;;; http://www.mozilla.org/MPL/
;;;
;;; Software distributed under the License is distributed on an "AS IS"
;;; basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
;;; License for the specific language governing rights and limitations
;;; under the License.
;;;
;;; The Original Code is SNARK.
;;; The Initial Developer of the Original Code is SRI International.
;;; Portions created by the Initial Developer are Copyright (C) 1981-2007.
;;; All Rights Reserved.
;;;
;;; Contributor(s): Mark E. Stickel <stickel@ai.sri.com>.

(in-package :snark)

(defun print-sort (sort &optional (stream *standard-output*) depth)
  (declare (ignore depth))
  (with-standard-io-syntax2
    (prin1 (sort-name sort) stream))
  sort)

(defun print-function-symbol (fn &optional (stream *standard-output*) depth)
  (declare (ignore depth))
  (with-standard-io-syntax2
    (prin1 (function-name fn) stream))
  fn)

(defun print-term3 (term &optional (stream *standard-output*) depth)
  ;; compound-appl, variable, function-symbol struct printer
  (declare (ignore depth))
  (print-term term *subst-for-printing* stream))

(defun print-term (term &optional subst (stream *standard-output*))
  ;; terms are printed by first converting them to lisp
  ;; by replacing compound structures by conses, function structures by symbols, etc.
  (let ((x (let-options ((variable-to-lisp-code nil)) (term-to-lisp term subst))))
    (with-standard-io-syntax2
      (cond
       ((variable-p x)
        ;; variables may not be changed by term-to-lisp, print them here
        (let ((num (variable-number x))
              (sort (variable-sort x)))
          (princ (first (variable-symbol-prefixes?)) stream)
          (cond
           ((not (top-sort? sort))
            (princ (sort-name sort) stream)
            (unless (eql 0 num)
              (princ num stream)))
           (t
            (mvlet (((:values i j) (floor num 6)))
              (princ (nth j '(x y z u v w)) stream)
              (unless (eql 0 i)
                (princ i stream)))))))
       ((and (consp x)					;override printing of 'a and #'a for (quote a) and (function a)
             (or (eq 'quote (first x)) (eq 'function (first x)))
             (rest x)
             (null (rest (rest x))))
        (princ #\( stream)
        (prin1 (first x) stream)
        (princ #\  stream)
        (prin1 (second x) stream)
        (princ #\) stream))
       (t
        (prin1 x stream))))
    term))

(defun print-row-term (term &optional subst (stream *standard-output*))
  (let ((term term))
    (when (print-row-length-limit?)
      (dereference
       term subst
       :if-compound-appl (when (and (eq *or* (heada term)) (< (print-row-length-limit?) (length (argsa term))))
                           (setf term (make-compound* *or* (nconc (firstn (argsa term) (print-row-length-limit?)) '(---)))))))
    (let ((*print-pretty2* (and (print-rows-prettily?) (print-row-wffs-prettily?))))
      (print-term term subst stream)))
  term)

(defmethod print-given-row (row)
  (case (print-rows-when-given?)
    ((nil)
     (when (eq :signal (print-rows-when-derived?))
       (comment)
       (princ #\|)))
    (:signal
     (comment)
     (princ #\|))
    (:indicate				;ttp 7/13/93
     (with-clock-on printing
       (terpri)
       (terpri-comment)
       (format t "*** Inferences from ~A *** " (row-name-or-number row))))
    (otherwise
     (with-clock-on printing
       (when (print-time-used?)
	 (print-incremental-time-used))
       (dotimes (dummy (- (case (print-rows-when-derived?)
                            ((:signal nil)
                             (print-given-row-lines-signalling?))
                            (otherwise
                             (print-given-row-lines-printing?)))
                          1))
	 (declare (ignorable dummy))
	 (terpri))
       (terpri)
       (print-row row :string "Infer from row ")
       (princ " ")
       (force-output))))
  row)

(defmethod print-derived-row (row)
  (case (print-rows-when-derived?)
    ((nil)
     )
    (:signal
     (comment)
     (princ #\+))
    #+ignore
    (:fact
     (when (let ((wff (row-wff row)))
             (dereference wff nil :if-compound (eq fact-relation (head wff))))
       (with-clock-on printing
         (when (print-time-used?)
           (print-incremental-time-used))
         (terpri)
         (print-row row)
         (princ " "))))
    (:indicate				;ttp 7/29/93
     (with-clock-on printing
       (comment)
       (princ " ")
       (princ (row-name-or-number row))))
    (otherwise
     (with-clock-on printing
       (when (print-time-used?)
	 (print-incremental-time-used))
       (terpri)
       (print-row row)
       (princ " "))))
  row)

(defun print-processed-row (row)
  (case (print-rows-when-processed?)
    ((nil :signal :indicate)
     )
    (otherwise
     (with-clock-on printing
       (when (print-time-used?)
	 (print-incremental-time-used))
       (terpri)
       (print-row row :string "Processing row ")
       (princ " "))))
  row)

(defvar *printing-deleted-messages* nil)

(defun print-deleted-wff (row msg)
  (case (print-rows-when-derived?)
    ((nil)
     )
    (:signal
     (comment)
     (princ (if (equal "deleted because agenda full" msg) #\d #\-)))
    #+ignore
    (:fact
     (when (let ((wff (row-wff row)))
             (dereference wff nil :if-compound (eq fact-relation (head wff))))
       (with-clock-on printing
         (terpri-comment)
         (format t "     ~A ~A" msg (row-name-or-number row)))))
    (otherwise
     (with-clock-on printing
       (cond
        ((equal *printing-deleted-messages* msg)
         (format t ",~A" (row-name-or-number row)))
        (t
         (terpri-comment)
         (format t "~A ~A" msg (row-name-or-number row))
         (setf *printing-deleted-messages* msg))))))
  row)

(defun print-unorientable-wff (equality-or-equivalence)
  (case (print-unorientable-rows?)
    ((nil :signal :indicate)
     )
    (otherwise
     (with-clock-on printing
       (warn "Could not orient ~A." equality-or-equivalence))))
  equality-or-equivalence)

(defun print-final-row (row)
  (let ((p (print-final-rows?)))
    (cond
     ((null p)
      )
     ((or (eq :signal p) (eq :indicate p))
      (comment)
      (princ #\.))
     (t
      (with-clock-on printing
        (terpri)
        (terpri)
        (princ "(Refutation")
        (print-ancestry row)
        (terpri)
        (princ ")")
        (cond
         ((eq :tptp p)
          (terpri)
          (terpri)
          (let ((answer (row-answer row)))
            (when (and (compound-p answer) (eq 'values (function-name (head answer))))
              (princ "SZS answers short")
              (dolist (term (args answer))
                (princ " ")
                (print-term-in-tptp-format term))
              (terpri)
              (terpri)))
          (princ "refutation.")
          (print-ancestry row :format :tptp)
          (terpri)
          (princ "end_refutation."))))))
    row))

(defun replace-rows-by-name-or-number (x)
  (cond
   ((consp x)
    (lcons (replace-rows-by-name-or-number (car x)) (replace-rows-by-name-or-number (cdr x)) x))
   ((row-p x)
    (row-name-or-number x))
   (t
    x)))

(defun print-row-reason (row)
  (with-standard-io-syntax2
    (prin1 (replace-rows-by-name-or-number (row-reason row))))
  nil)

(defun print-row3 (row *standard-output* depth)
  "this function is used in the defstruct for ROW to print rows."
  (declare (ignore depth))
  (let-options ((print-rows-shortened nil)
		(print-rows-prettily nil)
                (print-row-reasons nil)
		(print-row-answers nil)
		(print-row-constraints nil)
                (print-row-partitions nil))
    (print-row row)))

(defun print-row-length-limit1 (row)
  (let ((n1 (print-rows-shortened?)))
    (and n1
         (let* ((reason (row-reason row))
                (n2 (and (consp reason)
	                 (eq 'resolve (first reason))
	                 (row-p (third reason))
	                 (clause-p (row-wff (third reason)))
	                 (wff-length (row-wff (third reason))))))
           (if (numberp n1)
               (if n2 (min n1 n2) n1)
               n2)))))

(defun print-row (row &key (string "Row ") format)
  (when row
    (ecase format
      ((nil)
       (with-standard-io-syntax2
         (princ "(")
         (princ string)
         (prin1 (row-name-or-number row))
         (cond
          ((print-rows-prettily?)
           (terpri)
           (princ "   "))
          (t
           (princ " ")))
         (let-options ((print-row-length-limit (print-row-length-limit1 row)))
           (print-row-term
            (cond
             ((not (print-row-goals?))
              (prog->
                (map-atoms-in-wff-and-compose-result (row-wff row) ->* atom polarity)
                (declare (ignore polarity))
                (dereference
                 atom nil
                 :if-constant (if (proposition-magic-goal-p atom) true atom)
                 :if-compound (if (relation-magic-goal-p (head atom)) true atom))))
             (t
              (row-wff row)))))
         (when (print-row-reasons?)
           (cond
            ((print-rows-prettily?)
             (terpri)
             (princ "   "))
            (t
             (format t "~70T")))
           (print-row-reason row))
         (when (print-row-constraints?)
           (dolist (x (row-constraints row))
             (unless (eq true (cdr x))
               (terpri)
               (princ "   ")
               (princ (string-capitalize (car x)))
               (princ "-Constraint ")
               (print-row-term (cdr x)))))
         (when (print-row-answers?)
           (let ((answer (row-answer row)))
             (unless (eq false answer)
               (terpri)
               (princ "   Answer ")
               (print-row-term answer))))
         (when (and (use-partitions?) (print-row-partitions?))
           (terpri)
           (princ "   Partitions ")
           (prin1 (mapcar #'car (row-context row))))
         (princ ")")))
      (:tptp
       (print-row-in-tptp-format row))))
  row)

(defvar *propositional-abstraction-term-to-lisp* nil)

(defun term-to-lisp (term &optional subst)
  "Return a Lisp data structure for the given term."
  ;; returns (f a b c) for SNARK term f(a,b,c)
  ;; returns (list a b c) for SNARK term [a,b,c]
  ;;  use variable-p, variable-number, variable-sort
  ;;  sort information is invalid after SNARK is reinitialized
  (labels
    ((term-to-lisp (term)
       (dereference
	term subst
	:if-constant (let ((name (constant-name term)))
                       (cond
                        ((can-be-free-variable-name name)
                         (kwote name))
                        (t
                         name)))
	:if-variable (dolist (fun (mklist (variable-to-lisp-code?)) term)
                       (let ((v (funcall fun term)))
                         (unless (eq none v)
                           (return v))))
	:if-compound (let ((head (head term))
                           (args (args term)))
                       (cond
                        ((and *propositional-abstraction-term-to-lisp*
                              (not (function-logical-symbol-p head)))
                         (list (function-name head) (function-arity head)))
                        (t
                         (dolist (fun (function-to-lisp-code head) (cons (function-name head) (args-to-lisp args)))
                           (let ((v (funcall fun head args subst)))
                             (unless (eq none v)
                               (return v)))))))))
     (args-to-lisp (args)
       (lcons (term-to-lisp (first args)) (args-to-lisp (rest args)) args)))
    (term-to-lisp term)))

(defun cons-term-to-lisp (head args subst)
  ;; converts
  ;;   (a)        to ($$list a)
  ;;   (a b)      to ($$list a b)
  ;;   (a . b)    to ($$cons a b)
  ;;   (a b . c)  to ($$list* a b c)
  ;; when used as to-lisp-code for cons
  (cl:assert (eq *cons* head))
  (let* ((y (term-to-lisp (second args) subst))
         (x (term-to-lisp (first args) subst)))
    (cond
     ((null y)
      (list (function-name (find-symbol-table-entry '$$list :function :any)) x))
     ((atom y)
      (list (function-name head) x y))
     (t
      (let ((v (first y)) list*)
        (cond
         ((eq v (function-name (find-symbol-table-entry '$$list :function :any)))
          (list* v x (rest y)))
         ((or (eq v (setf list* (function-name (find-symbol-table-entry '$$list* :function :any))))
              (eq v (function-name head)))
          (list* list* x (rest y)))
         (t
          (list (function-name head) x y))))))))

(defun quant-compound-to-lisp (head args subst)
  (list (function-name head)
        (mapcar (lambda (var-spec)
                  (if (variable-p var-spec)
                      (term-to-lisp var-spec subst)
                      (mapcar #'(lambda (x) (term-to-lisp x subst)) var-spec)))
                (first args))
        (term-to-lisp (second args) subst)))

(defun row-sorts (row &optional sorts)
  (prog->
   (map-terms-in-wff (row-wff row) ->* term polarity)
   (declare (ignore polarity))
   (let ((sort (term-sort term)))
     (unless (top-sort? sort)
       (pushnew (term-sort term) sorts :test #'same-sort?))))
  sorts)

(defun derivation-sorts (row)
  (let ((sorts nil))
    (dolist (row (row-ancestry row))
      (setf sorts (row-sorts row sorts)))
    sorts))

(defun subsort-forms (sorts)
  (let ((result nil))
    (dotails (l sorts)
      (let ((sort1 (first l)))
	(dolist (sort2 (rest l))
	  (cond
	    ((subsort? sort1 sort2)
	     (push `(subsort ,(sort-name sort1) ,(sort-name sort2)) result))
	    ((subsort? sort2 sort1)
	     (push `(subsort ,(sort-name sort2) ,(sort-name sort1)) result))))))
    result))

(defun derivation-subsort-forms (row)
  (subsort-forms (derivation-sorts row)))

;;; output.lisp EOF
